@@ -19,13 +19,20 @@
 
 from io import StringIO
 import re
+import sys
 from typing import Iterable, List, Union, Callable, Optional, TextIO, Iterator
+from urllib.parse import urljoin
 from warnings import warn
 
+from debian.changelog import Version
+
 from .reformatting import (
     Editor,
     )
 
+from . import __version__
+
+DEFAULT_USER_AGENT = 'debmutate/%s' % '.'.join([str(x) for x in __version__])
 
 DEFAULT_VERSION: int = 4
 
@@ -78,6 +85,22 @@ class WatchFile(object):
                 return value
         raise KeyError(name)
 
+    def set_option(self, name, newvalue=None):
+        if newvalue is None:
+            nv = name
+        else:
+            nv = '%s=%s' % (name, newvalue)
+        for i, option in enumerate(self.options):
+            try:
+                key, value = option.split('=', 1)
+            except ValueError:
+                key = option
+                value = None
+            if key == value:
+                self.options[i] = nv
+                return
+        self.options.append(nv)
+
     def del_option(self, name):
         for i, option in enumerate(self.options):
             try:
@@ -123,6 +146,69 @@ class WatchFile(object):
             f.write('\n')
 
 
+def parse_subst_expr(vm):
+    if vm[0] != 's':
+        raise InvalidUVersionMangle(vm, 'not a substitution regex')
+    parts = re.split(r'(?<!\\)' + vm[1], vm)
+    if len(parts) < 3:
+        raise InvalidUVersionMangle(vm)
+    pattern = parts[1]
+    replacement = parts[2]
+    flags = parts[3]
+    return (pattern, replacement, flags)
+
+
+def apply_subst_expr(vm: str, orig: str) -> str:
+    (pattern, replacement, flags) = parse_subst_expr(vm)
+    # TODO(jelmer): Handle flags
+    return re.sub(pattern, replacement, orig)
+
+
+def apply_url_mangle(expr: str, orig: str) -> str:
+    return apply_subst_expr(expr, orig)
+
+
+class Release(object):
+    """A discovered release."""
+
+    def __init__(self, version, url, pgpsigurl=None):
+        self.version = version
+        self.url = url
+        self.pgpsigurl = pgpsigurl
+
+    def __lt__(self, other):
+        if type(self) != type(other):
+            raise TypeError(other)
+        return Version(self.version) < Version(other.version)
+
+    def __repr__(self):
+        return "%s(%r, %r, pgpsigurl=%r)" % (
+            type(self).__name__, self.version, self.url,
+            self.pgpsigurl)
+
+
+def html_search(body, matching_pattern):
+    from bs4 import BeautifulSoup
+    soup = BeautifulSoup(body, 'html.parser')
+    for a in soup.find_all('a'):
+        href = a.attrs.get('href')
+        if not href:
+            continue
+        m = re.fullmatch(matching_pattern, href)
+        if m:
+            yield m
+
+
+def plain_search(body, matching_pattern):
+    return re.finditer(matching_pattern, body)
+
+
+searchers = {
+    'plain': plain_search,
+    'html': html_search,
+    }
+
+
 class Watch(object):
 
     def __init__(self, url: str, matching_pattern: Optional[str] = None,
@@ -141,10 +227,8 @@ class Watch(object):
             vm = self.get_option('uversionmangle')
         except KeyError:
             return version
-        m = re.fullmatch('^s/(.*)/(.*[^/])(/[a-zA-Z]*)?$', vm)
-        if not m:
-            raise InvalidUVersionMangle(vm)
-        return re.sub(m.group(1), m.group(2).replace('$', '\\'), version)
+        (pattern, replacement, flags) = parse_subst_expr(vm)
+        return re.sub(pattern, replacement.replace('$', '\\'), version)
 
     def get_option(self, name):
         for option in self.options:
@@ -157,6 +241,29 @@ class Watch(object):
                 return value
         raise KeyError(name)
 
+    def has_option(self, name):
+        try:
+            self.get_option(name)
+        except KeyError:
+            return False
+        return True
+
+    def set_option(self, name, newvalue=None):
+        if newvalue is None:
+            nv = name
+        else:
+            nv = '%s=%s' % (name, newvalue)
+        for i, option in enumerate(self.options):
+            try:
+                key, value = option.split('=', 1)
+            except ValueError:
+                key = option
+                value = None
+            if key == value:
+                self.options[i] = nv
+                return
+        self.options.append(nv)
+
     def del_option(self, name):
         for i, option in enumerate(self.options):
             try:
@@ -190,6 +297,30 @@ class Watch(object):
             package = package()
         return self.url.replace('@PACKAGE@', package)
 
+    def discover(self, package):
+        from urllib.request import urlopen, Request
+        url = self.format_url(package)
+        try:
+            user_agent = self.get_option('user-agent')
+        except KeyError:
+            user_agent = DEFAULT_USER_AGENT
+        try:
+            searchmode = self.get_option('searchmode')
+        except KeyError:
+            searchmode = 'html'
+        req = Request(url, headers={'User-Agent': user_agent})
+        resp = urlopen(req)
+        for m in searchers[searchmode](resp.read(), self.matching_pattern):
+            # TODO(jelmer): Apply uversionmangle
+            full_url = urljoin(url, m.group(0))
+            try:
+                pgpsigurlmangle = self.get_option('pgpsigurlmangle')
+            except KeyError:
+                pgpsigurl = None
+            else:
+                pgpsigurl = apply_url_mangle(pgpsigurlmangle, full_url)
+            yield Release(m.group(1), full_url, pgpsigurl=pgpsigurl)
+
 
 class MissingVersion(Exception):
     """The version= line is missing."""
@@ -301,3 +432,22 @@ class WatchEditor(Editor):
         nf = StringIO()
         parsed.dump(nf)
         return nf.getvalue()
+
+
+def uscan(wf, package):
+    for entry in wf.entries:
+        for d in entry.discover(package):
+            print('%s' % d)
+
+
+def main(argv):
+    with open('debian/watch', 'r') as f:
+        wf = parse_watch_file(f)
+    from debian.deb822 import Deb822
+    with open('debian/control', 'r') as f:
+        source = Deb822(f)
+    uscan(wf, source['Source'])
+
+
+if __name__ == '__main__':
+    sys.exit(main(sys.argv))
